from typing import List, Union, Optional

from discord import (
    User,
    Embed,
    AllowedMentions,
    InvalidArgument,
    Guild,
    NotFound,
    Member,
    Object,
    MessageFlags,
    File,
)
from discord.state import ConnectionState
from discord.abc import Messageable

from enum import IntEnum

from .utils import _get_components_json
from .component import Component, ActionRow, Button, Select
from .dpy_overrides import ComponentMessage


__all__ = ("Interaction", "InteractionEventType")


class InteractionEventType(IntEnum):
    button_click = 2
    select_option = 3


class Interaction:
    def __init__(
        self,
        *,
        state: ConnectionState,
        client: "DiscordComponents",
        raw_data: dict,
    ):
        self.state: ConnectionState = state
        self.client: "DiscordComponents" = client

        self.interaction_id: int = int(raw_data["id"])
        self.interaction_token: str = raw_data["token"]

        self.custom_id: str = raw_data["data"]["custom_id"]
        self.values: List[str] = raw_data["data"].get("values", [])
        self.component_type: int = raw_data["data"]["component_type"]

        self.channel_id: int = int(raw_data["channel_id"])
        self.guild_id: int = raw_data.get("guild_id")
        if self.guild_id is not None:
            self.guild_id = int(self.guild_id)

        if self.guild:
            self.user: Union[User, Member] = Member(
                state=state, guild=self.guild, data=raw_data["member"]
            )
        elif raw_data.get("member"):
            self.user: Union[User, Member] = User(state=state, data=raw_data["member"]["user"])
        else:
            self.user: Union[User, Member] = User(state=state, data=raw_data["user"])
        self.author: Union[User, Member] = self.user

        self.message: Union[ComponentMessage, dict] = ComponentMessage(
            state=state,
            channel=self.channel,
            data=raw_data["message"],
            ephemeral=raw_data["message"].get("flags") == 64,
        )
        self.component: Component = self.message.get_component(custom_id=self.custom_id)

        self.raw_data: dict = raw_data
        self.responded: bool = False
        self.deferred: bool = False

        self._deferred_hidden = False
        self._deferred_edit_origin = False

    @property
    def channel(self) -> Optional[Messageable]:
        return self.state.get_channel(self.channel_id)

    @property
    def guild(self) -> Optional[Guild]:
        return self.state._get_guild(self.guild_id)

    async def defer(self, ephemeral: bool = True, edit_origin: bool = False):
        if self.deferred or self.responded:
            return

        await self.respond(type=5 if not edit_origin else 6, ephemeral=ephemeral)

        if ephemeral:
            self._deferred_hidden = True
        self._deferred_edit_origin = edit_origin

    async def respond(
        self,
        *,
        type: int = 4,
        content: str = None,
        embed: Embed = None,
        embeds: List[Embed] = None,
        suppress: bool = None,
        file: File = None,
        files: List[File] = None,
        allowed_mentions: AllowedMentions = None,
        tts: bool = False,
        ephemeral: bool = True,
        components: List[Union[ActionRow, Component, List[Component]]] = None,
    ) -> Optional[Union[ComponentMessage, dict]]:
        if self.responded:
            return

        state = self.state
        data = {"tts": tts}

        if ephemeral:
            data["flags"] = 64

        if content is not None:
            data["content"] = str(content)

        if embed is not None and embeds is not None:
            raise InvalidArgument("cannot pass both embed and embeds parameter")

        if embed is not None:
            data["embeds"] = [embed.to_dict()]

        if embeds is not None:
            if len(embeds) > 10:
                raise InvalidArgument("embeds parameter must be a list of up to 10 elements")
            data["embeds"] = [embed.to_dict() for embed in embeds]

        if suppress is not None:
            flags = MessageFlags._from_value(
                self.message.flags if isinstance(self.message, ComponentMessage) else 0
            )
            flags.suppress_embeds = suppress
            data["flags"] = flags.value

        if allowed_mentions is not None:
            if state.allowed_mentions is not None:
                data["allowed_mentions"] = state.allowed_mentions.merge(allowed_mentions).to_dict()
            else:
                data["allowed_mentions"] = allowed_mentions.to_dict()
        else:
            data["allowed_mentions"] = state.allowed_mentions and state.allowed_mentions.to_dict()

        if components is not None:
            data["components"] = _get_components_json(components)

        if not self.deferred:
            data = {"type": type, "data": data}

        if file is not None and files is not None:
            raise InvalidArgument("cannot pass both file and files parameter to send()")
        elif files is not None:
            if len(files) > 10:
                raise InvalidArgument("files parameter must be a list of up to 10 elements")
        if file is not None:
            files = [file]

        try:
            if self.deferred:
                res = await self.client.http.edit_response(
                    interaction_token=self.interaction_token, data=data, files=files
                )
            else:
                res = await self.client.http.initial_response(
                    interaction_id=self.interaction_id,
                    interaction_token=self.interaction_token,
                    data=data,
                    files=files,
                )

            if type in (4, 7):
                self.responded = True
            else:
                self.deferred = True
        except NotFound as e:
            self.responded = True
            raise NotFound(
                e.response,
                "Interaction is unknown (you have already responded to the interaction or responding took too long)",
            ) from None

        if type in (4, 7) and isinstance(res, dict):
            return ComponentMessage(
                state=state,
                data=res,
                channel=self.channel or Object(id=self.channel_id),
                ephemeral=res.get("flags") == 64,
            )
        else:
            return res

    async def send(
        self,
        content: str = None,
        embed: Embed = None,
        embeds: List[Embed] = None,
        file: File = None,
        files: List[File] = None,
        allowed_mentions: AllowedMentions = None,
        tts: bool = False,
        ephemeral: bool = True,
        components: List[Union[ActionRow, Component, List[Component]]] = None,
        delete_after: float = None,
    ) -> Optional[Union[ComponentMessage, dict]]:
        await self.defer(ephemeral=ephemeral)
        res = await self.respond(
            type=4,
            content=content,
            embed=embed,
            embeds=embeds,
            file=file,
            files=files,
            allowed_mentions=allowed_mentions,
            tts=tts,
            ephemeral=ephemeral,
            components=components,
        )

        if delete_after is None:
            return res
        elif res is not None:
            await res.delete(delay=delete_after)

    async def edit_origin(
        self,
        content: str = None,
        embed: Embed = None,
        embeds: List[Embed] = None,
        suppress: bool = None,
        file: File = None,
        files: List[File] = None,
        allowed_mentions: AllowedMentions = None,
        components: List[Union[ActionRow, Component, List[Component]]] = None,
        delete_after: float = None,
    ):
        await self.defer(edit_origin=True)
        res = await self.respond(
            type=7,
            content=content,
            embed=embed,
            embeds=embeds,
            suppress=suppress,
            file=file,
            files=files,
            allowed_mentions=allowed_mentions,
            components=components,
        )

        if delete_after is None:
            return res
        elif res is not None:
            await res.delete(delay=delete_after)

    async def disable_components(self) -> None:
        await self.edit_origin(
            components=[row.disable_components() for row in self.message.components],
        )



from typing import Optional, List, Union

from discord import (
    Message,
    Embed,
    Attachment,
    AllowedMentions,
    InvalidArgument,
    File,
    MessageFlags,
)
from discord.http import Route, HTTPClient
from discord.abc import Messageable, Snowflake
from discord.ext.commands import Context

from .utils import _get_components_json, _form_files
from .component import _get_component_type, ActionRow, Component

__all__ = ("ComponentMessage",)


class ComponentMessage(Message):
    __slots__ = tuple(list(Message.__slots__) + ["components", "ephemeral"])

    def __init__(self, *, state, channel, data, ephemeral=False):
        super().__init__(state=state, channel=channel, data=data)
        self.ephemeral = ephemeral

        components = []
        for i in data["components"]:
            components.append(ActionRow())
            for j in i["components"]:
                components[-1].append(_get_component_type(j["type"]).from_json(j))
        self.components: List[ActionRow] = components

    def get_component(self, custom_id: str) -> Optional[Component]:
        for row in self.components:
            for component in row.components:
                if component.custom_id == custom_id:
                    return component

    async def disable_components(self) -> None:
        await self.edit(
            components=[row.disable_components() for row in self.components],
        )

    async def edit(
        self,
        content: Optional[str] = None,
        embed: Optional[Embed] = None,
        embeds: List[Embed] = None,
        suppress: bool = None,
        attachments: List[Attachment] = None,
        delete_after: Optional[float] = None,
        allowed_mentions: Optional[AllowedMentions] = None,
        components: List[Union[ActionRow, Component, List[Component]]] = None,
        **fields,
    ):
        if self.ephemeral:
            return

        state = self._state
        data = {}

        if content is not None:
            data["content"] = content

        if embed is not None and embeds is not None:
            raise InvalidArgument("cannot pass both embed and embeds parameter to edit()")

        if embed is not None:
            data["embeds"] = [embed.to_dict()]

        if embeds is not None:
            data["embeds"] = [e.to_dict() for e in embeds]

        if suppress is not None:
            flags = MessageFlags._from_value(0)
            flags.suppress_embeds = True
            data["flags"] = flags.value

        if allowed_mentions is None:
            if state.allowed_mentions is not None and self.author.id == self._state.self_id:
                data["allowed_mentions"] = state.allowed_mentions.to_dict()
        else:
            if state.allowed_mentions is not None:
                data["allowed_mentions"] = state.allowed_mentions.merge(allowed_mentions).to_dict()
            else:
                data["allowed_mentions"] = allowed_mentions.to_dict()

        if attachments is not None:
            data["attachments"] = [a.to_dict() for a in attachments]

        if components is not None:
            data["components"] = _get_components_json(components)

        if data:
            await state.http.request(
                Route(
                    "PATCH",
                    "/channels/{channel_id}/messages/{message_id}",
                    channel_id=self.channel.id,
                    message_id=self.id,
                ),
                json=data,
            )

        if delete_after is not None:
            await self.delete(delay=delete_after)

    async def delete(self, *args, **kwargs):
        if self.ephemeral:
            return

        return await super().delete(*args, **kwargs)


def new_override(cls, *args, **kwargs):
    if isinstance(cls, Message):
        return object.__new__(ComponentMessage)
    else:
        return object.__new__(cls)


Message.__new__ = new_override


def send_files(
    self,
    channel_id: Snowflake,
    *,
    files,
    content=None,
    tts=False,
    embed=None,
    embeds=None,
    stickers=None,
    nonce=None,
    allowed_mentions=None,
    message_reference=None,
    components=None,
):
    data = {"tts": tts}
    if content is not None:
        data["content"] = content
    if embed is not None:
        data["embeds"] = [embed]
    if embeds is not None:
        data["embeds"] = embeds
    if nonce is not None:
        data["nonce"] = nonce
    if allowed_mentions is not None:
        data["allowed_mentions"] = allowed_mentions
    if message_reference is not None:
        data["message_reference"] = message_reference
    if stickers is not None:
        data["sticker_ids"] = stickers
    if components is not None:
        data["components"] = components

    form = _form_files(data, files, use_form=False)
    return self.request(
        Route("POST", "/channels/{channel_id}/messages", channel_id=channel_id),
        form=form,
        files=files,
    )


def send_message(
    self,
    channel_id,
    content,
    *,
    tts=False,
    embed=None,
    embeds=None,
    nonce=None,
    allowed_mentions=None,
    message_reference=None,
    stickers=None,
    components=None,
):
    payload = {"tts": tts}

    if content is not None:
        payload["content"] = content

    if embed is not None:
        payload["embeds"] = [embed]

    if embeds is not None:
        payload["embeds"] = embeds

    if nonce is not None:
        payload["nonce"] = nonce

    if allowed_mentions is not None:
        payload["allowed_mentions"] = allowed_mentions

    if message_reference is not None:
        payload["message_reference"] = message_reference

    if stickers is not None:
        payload["sticker_ids"] = stickers

    if components is not None:
        payload["components"] = components

    return self.request(
        Route("POST", "/channels/{channel_id}/messages", channel_id=channel_id),
        json=payload,
    )


HTTPClient.send_files = send_files
HTTPClient.send_message = send_message


async def send(
    self,
    content=None,
    *,
    tts=False,
    embed=None,
    embeds=None,
    file=None,
    files=None,
    stickers=None,
    delete_after=None,
    nonce=None,
    allowed_mentions=None,
    reference=None,
    mention_author=None,
    components=None,
):
    state = self._state
    channel = await self._get_channel()
    content = str(content) if content is not None else None

    if embed is not None and embeds is not None:
        raise InvalidArgument("cannot pass both embed and embeds parameter to send()")

    if embed is not None:
        embeds = [embed.to_dict()]

    elif embeds is not None:
        if len(embeds) > 10:
            raise InvalidArgument("embeds parameter must be a list of up to 10 elements")
        embeds = [embed.to_dict() for embed in embeds]

    if stickers is not None:
        stickers = [sticker.id for sticker in stickers]

    if allowed_mentions is not None:
        if state.allowed_mentions is not None:
            allowed_mentions = state.allowed_mentions.merge(allowed_mentions).to_dict()
        else:
            allowed_mentions = allowed_mentions.to_dict()
    else:
        allowed_mentions = state.allowed_mentions and state.allowed_mentions.to_dict()

    if mention_author is not None:
        allowed_mentions = allowed_mentions or AllowedMentions().to_dict()
        allowed_mentions["replied_user"] = bool(mention_author)

    if reference is not None:
        try:
            reference = reference.to_message_reference_dict()
        except AttributeError:
            raise InvalidArgument(
                "reference parameter must be Message or MessageReference"
            ) from None

    if components is not None:
        components = _get_components_json(components)

    if file is not None and files is not None:
        raise InvalidArgument("cannot pass both file and files parameter to send()")

    if file is not None:
        if not isinstance(file, File):
            raise InvalidArgument("file parameter must be File")

        try:
            data = await state.http.send_files(
                channel.id,
                files=[file],
                allowed_mentions=allowed_mentions,
                content=content,
                tts=tts,
                embed=embed,
                embeds=embeds,
                nonce=nonce,
                message_reference=reference,
                stickers=stickers,
                components=components,
            )
        finally:
            file.close()

    elif files is not None:
        if len(files) > 10:
            raise InvalidArgument("files parameter must be a list of up to 10 elements")
        elif not all(isinstance(file, File) for file in files):
            raise InvalidArgument("files parameter must be a list of File")

        try:
            data = await state.http.send_files(
                channel.id,
                files=files,
                content=content,
                tts=tts,
                embed=embed,
                embeds=embeds,
                nonce=nonce,
                allowed_mentions=allowed_mentions,
                message_reference=reference,
                stickers=stickers,
                components=components,
            )
        finally:
            for f in files:
                f.close()
    else:
        data = await state.http.send_message(
            channel.id,
            content,
            tts=tts,
            embed=embed,
            embeds=embeds,
            nonce=nonce,
            allowed_mentions=allowed_mentions,
            message_reference=reference,
            stickers=stickers,
            components=components,
        )

    ret = ComponentMessage(state=state, channel=channel, data=data)
    if delete_after is not None:
        await ret.delete(delay=delete_after)
    return ret


async def send_override(context_or_channel, *args, **kwargs):
    if isinstance(context_or_channel, Context):
        channel = context_or_channel.channel
    else:
        channel = context_or_channel

    return await send(channel, *args, **kwargs)


async def fetch_message(context_or_channel, id: int):
    if isinstance(context_or_channel, Context):
        channel = context_or_channel.channel
    else:
        channel = context_or_channel

    state = channel._state
    data = await state.http.get_message(channel.id, id)
    return ComponentMessage(state=state, channel=channel, data=data)


Messageable.send = send_override
Messageable.fetch_message = fetch_message



from typing import List

from discord import Client, File
from discord.http import Route

from .utils import _form_files


__all__ = ("HTTPClient",)


class HTTPClient:
    def __init__(self, bot: Client):
        self.bot = bot

    def edit_response(
        self, interaction_token: str, data: dict, files: List[File] = None
    ):
        route = Route(
            "PATCH",
            f"/webhooks/{self.bot.user.id}/{interaction_token}/messages/@original",
        )

        if files is not None:
            return self.bot.http.request(
                route, data=_form_files(data, files), files=files
            )
        else:
            return self.bot.http.request(
                route,
                json=data,
            )

    def initial_response(
        self,
        interaction_id: int,
        interaction_token: str,
        data: dict,
        files: List[File] = None,
    ):
        route = Route(
            "POST",
            f"/interactions/{interaction_id}/{interaction_token}/callback",
        )

        if files is not None:
            return self.bot.http.request(
                route, data=_form_files(data, files), files=files
            )
        else:
            return self.bot.http.request(
                route,
                json=data,
            )



from discord import Message, Guild, TextChannel, User

from discord_components.interaction import Interaction
from discord_components.component import Component, Button, SelectOption


__all__ = (
    "message_filter",
    "component_filter",
    "guild_filter",
    "channel_filter",
    "user_filter",
)


def message_filter(message: Message, ephemeral: bool = False):
    def _filter(interaction: Interaction):
        if ephemeral or isinstance(interaction.message, dict):
            return False

        return interaction.message.id == message.id

    return _filter


def component_filter(component: Component):
    def _filter(interaction: Interaction):
        return interaction.custom_id == component.id

    return _filter


def guild_filter(guild: Guild):
    def _filter(interaction: Interaction):
        return interaction.guild_id == guild.id

    return _filter


def channel_filter(channel: TextChannel):
    def _filter(interaction: Interaction):
        return interaction.channel_id == channel.id

    return _filter


def user_filter(user: User):
    def _filter(interaction: Interaction):
        return interaction.user.id == user.id

    return _filter



from typing import Union

from discord import (
    Client,
    Message,
    User,
    Guild,
)
from discord.ext.commands import Bot
from discord.abc import Messageable

from .component import Component
from .http import HTTPClient
from .interaction import Interaction, InteractionEventType

from .ext.filters import *

__all__ = ("DiscordComponents", "ComponentsClient", "ComponentsBot")


class DiscordComponents:
    def __init__(
        self,
        bot: Union[Bot, Client],
    ):
        self.bot = bot
        bot.components_manager = self

        self.http = HTTPClient(bot=bot)
        self._components_callback = {}

        if isinstance(self.bot, Bot):
            self.bot.add_listener(self.on_socket_response, name="on_socket_response")
        else:
            self.bot.on_socket_response = self.on_socket_response

    async def on_socket_response(self, res):
        if (res["t"] != "INTERACTION_CREATE") or (res["d"]["type"] != 3):
            return

        if res["d"]["message"].get("message_reference") and not res["d"]["message"][
            "message_reference"
        ].get("channel_id"):
            res["d"]["message"]["message_reference"] = res["d"]["channel_id"]

        interaction = self._get_interaction(res)
        self.bot.dispatch(f"raw_interaction", res["d"])
        self.bot.dispatch("interaction", interaction)

        if self._components_callback.get(interaction.custom_id):
            callback_info = self._components_callback[interaction.custom_id]
            if callback_info["uses"] == 0:
                del self._components_callback[interaction.custom_id]
                return

            if callback_info["uses"] is not None:
                self._components_callback[interaction.custom_id]["uses"] -= 1
            if not callback_info["filter"](interaction):
                return

            await self._components_callback[interaction.custom_id]["callback"](interaction)

        for _type in InteractionEventType:
            if _type.value == res["d"]["data"]["component_type"]:
                self.bot.dispatch(f"raw_{_type.name}", res["d"])
                self.bot.dispatch(_type.name, interaction)
                break

    def _get_interaction(self, json: dict):
        ctx = Interaction(
            state=self.bot._connection,
            client=self,
            raw_data=json["d"],
        )
        return ctx

    async def wait_for(
        self,
        event: str,
        *,
        message: Message = None,
        component: Component = None,
        ephemeral: bool = False,
        guild: Guild = None,
        channel: Messageable = None,
        user: User = None,
        timeout: float = None,
    ):
        check_list = []
        if message is not None:
            check_list.append(message_filter(message, ephemeral))
        if component is not None:
            check_list.append(component_filter(component))
        if guild is not None:
            check_list.append(guild_filter(guild))
        if channel is not None:
            check_list.append(channel_filter(channel))
        if user is not None:
            check_list.append(user_filter(user))

        def check(interaction: Interaction):
            for i in check_list:
                if not i(interaction):
                    return False
            return True

        return await self.bot.wait_for(event, check=check, timeout=timeout)

    def add_callback(self, component: Component, callback, *, uses: int = None, filter=None):
        self._components_callback[component.custom_id] = {
            "callback": callback,
            "uses": uses,
            "filter": filter or (lambda x: True),
        }
        return component


class ComponentsClient(Client):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.components_manager = DiscordComponents(self)


class ComponentsBot(Bot):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.components_manager = DiscordComponents(self)



from typing import List, Union, Optional

from discord import File

from aiohttp import FormData
from json import dumps

from .component import ActionRow, Component


__all__ = ("_get_components_json",)


def _get_components_json(
    components: List[Union[ActionRow, Component, List[Component]]] = None
) -> Optional[List[dict]]:
    if components is None:
        return None

    for i in range(len(components)):
        if isinstance(components[i], list):
            components[i] = ActionRow(*components[i])
        elif not isinstance(components[i], ActionRow):
            components[i] = ActionRow(components[i])

    lines = components
    return [row.to_dict() for row in lines] if lines else []


def _form_files(
    data: dict, files: List[File] = None, use_form: bool = True
) -> Union[FormData, List[dict]]:
    if use_form:
        form = FormData()
        form.add_field("payload_json", dumps(data))
        for i in range(len(files)):
            form.add_field(
                f"file{i if len(files) > 1 else ''}",
                files[i].fp,
                filename=files[i].filename,
                content_type="application/octet-stream",
            )
    else:
        form = [{"name": "payload_json", "value": dumps(data)}]
        for i in range(len(files)):
            form.append(
                {
                    "name": f"file{i if len(files) > 1 else ''}",
                    "value": files[i].fp,
                    "filename": files[i].filename,
                    "content_type": "application/octet-stream",
                }
            )
    return form



import re
from codecs import open
from os import path, environ

from setuptools import setup

PACKAGE_NAME = "discord_componentsss"
HERE = path.abspath(path.dirname(__file__))

with open("README.md", "r", encoding="utf-8") as f:
    README = f.read()

try:
    VERSION = (
        environ["TRAVIS_TAG"].lstrip("v")
        if environ["TRAVIS"] == "true"
        else environ["VERSION_NUMBER"]
    )
except KeyError:
    with open(path.join(HERE, PACKAGE_NAME, "const.py"), encoding="utf-8") as fp:
        VERSION = re.search('__version__ = "([^"]+)"', fp.read()).group(1)

extras = {
    "lint": ["black", "flake8", "isort"],
    "readthedocs": ["sphinx", "sphinx-rtd-theme"],
}
extras["lint"] += extras["readthedocs"]
extras["dev"] = extras["lint"] + extras["readthedocs"]


setup(
    name=PACKAGE_NAME,
    version="3.0.0",
    author="kiki7000",
    author_email="devkiki7000@gmail.com",
    description="An unofficial library for discord components.",
    extras_require=extras,
    include_package_data=True,
    license="MIT License",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/kiki7000/discord.py-components",
    packages=["discord_componentsss", "discord_componentsss.ext"],
    python_requires=">=3.6",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Internet",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Libraries",
        "Topic :: Utilities",
    ],
)



from typing import Optional, Union, List, Iterable

from discord import PartialEmoji, Emoji, InvalidArgument

from uuid import uuid1
from enum import IntEnum

__all__ = (
    "Component",
    "ButtonStyle",
    "Button",
    "Select",
    "SelectOption",
    "ActionRow",
    "_get_component_type",
)


def _get_partial_emoji(emoji: Union[Emoji, PartialEmoji, str]) -> PartialEmoji:
    if isinstance(emoji, Emoji):
        return PartialEmoji(name=emoji.name, animated=emoji.animated, id=emoji.id)
    elif isinstance(emoji, PartialEmoji):
        return emoji
    elif isinstance(emoji, str):
        return PartialEmoji(name=emoji)


class Component:
    def to_dict(self) -> dict:
        raise NotImplementedError

    @classmethod
    def from_json(cls, data: dict):
        raise NotImplementedError


class SelectOption(Component):
    __slots__ = ("_label", "_value", "_emoji", "_description", "_default")

    def __init__(
        self,
        *,
        label: str,
        value: str,
        emoji: Union[Emoji, PartialEmoji, str] = None,
        description: str = None,
        default: bool = False,
    ):
        self._label = label
        self._value = value
        self._description = description
        self._default = default

        if emoji is not None:
            self.emoji = _get_partial_emoji(emoji)
        else:
            self._emoji = None

    def to_dict(self) -> dict:
        data = {
            "label": self.label,
            "value": self.value,
            "description": self.description,
            "default": self.default,
        }
        if self.emoji is not None:
            data["emoji"] = self.emoji.to_dict()
        return data

    @property
    def label(self) -> str:
        return self._label

    @property
    def value(self) -> str:
        return self._value

    @property
    def emoji(self) -> Optional[PartialEmoji]:
        return self._emoji

    @property
    def description(self) -> str:
        return self._description

    @property
    def default(self) -> bool:
        return self._default

    @label.setter
    def label(self, value: str):
        if not len(value):
            raise InvalidArgument("Label must not be empty.")

        self._label = value

    @value.setter
    def value(self, value: str):
        self._value = value

    @emoji.setter
    def emoji(self, emoji: Union[Emoji, PartialEmoji, str]):
        self._emoji = _get_partial_emoji(emoji)

    @description.setter
    def description(self, value: str):
        self._description = value

    @default.setter
    def default(self, value: bool):
        self._default = value

    def set_label(self, value: str):
        self.label = value

    def set_value(self, value: str):
        self.value = value

    def set_emoji(self, emoji: Union[Emoji, PartialEmoji, str]):
        self.emoji = emoji

    def set_description(self, value: str):
        self.description = value

    def set_default(self, value: bool):
        self.default = value

    @classmethod
    def from_json(cls, data: dict):
        emoji = data.get("emoji")
        return cls(
            label=data.get("label"),
            value=data.get("value"),
            emoji=PartialEmoji(
                name=emoji["name"],
                animated=emoji.get("animated", False),
                id=emoji.get("id"),
            )
            if emoji
            else None,
            description=data.get("description"),
            default=data.get("default", False),
        )


class Select(Component):
    __slots__ = (
        "_id",
        "_options",
        "_placeholder",
        "_min_values",
        "_max_values",
        "_disabled",
    )

    def __init__(
        self,
        *,
        options: List[SelectOption],
        id: str = None,
        custom_id: str = None,
        placeholder: str = None,
        min_values: int = 1,
        max_values: int = 1,
        disabled: bool = False,
    ):
        if (not len(options)) or (len(options) > 25):
            raise InvalidArgument("Options length should be between 1 and 25.")

        self._id = id or custom_id or str(uuid1())
        self._options = options
        self._placeholder = placeholder
        self._min_values = min_values
        self._max_values = max_values
        self._disabled = disabled

    def to_dict(self) -> dict:
        return {
            "type": 3,
            "options": list(map(lambda option: option.to_dict(), self.options)),
            "custom_id": self.id,
            "placeholder": self.placeholder,
            "min_values": self.min_values,
            "max_values": self.max_values,
            "disabled": self.disabled,
        }

    @property
    def id(self) -> str:
        return self._id

    @property
    def custom_id(self) -> str:
        return self._id

    @property
    def options(self) -> List[SelectOption]:
        return self._options

    @property
    def placeholder(self) -> str:
        return self._placeholder

    @property
    def min_values(self) -> int:
        return self._min_values

    @property
    def max_values(self) -> int:
        return self._max_values

    @property
    def disabled(self) -> bool:
        return self._disabled

    @id.setter
    def id(self, value: str):
        self._id = value

    @custom_id.setter
    def custom_id(self, value: str):
        self._id = value

    @options.setter
    def options(self, value: List[SelectOption]):
        if (not len(value)) or (len(value) > 25):
            raise InvalidArgument("Options length should be between 1 and 25.")

        self._options = value

    @placeholder.setter
    def placeholder(self, value: str):
        self._placeholder = value

    @min_values.setter
    def min_values(self, value: int):
        self._min_values = value

    @max_values.setter
    def max_values(self, value: int):
        self._max_values = value

    @disabled.setter
    def disabled(self, value: bool):
        self._disabled = value

    def set_id(self, value: str):
        self.id = value

    def set_custom_id(self, value: str):
        self.custom_id = value

    def set_options(self, value: List[SelectOption]):
        self.options = value

    def set_placeholder(self, value: str):
        self.placeholder = value

    def set_min_values(self, value: int):
        self.min_values = value

    def set_max_values(self, value: int):
        self.max_values = value

    def set_disabled(self, value: bool):
        self.disabled = value

    @classmethod
    def from_json(cls, data: dict):
        return cls(
            id=data.get("custom_id"),
            options=list(map(lambda x: SelectOption.from_json(x), data.get("options"))),
            placeholder=data.get("placeholder"),
            min_values=data.get("min_values"),
            max_values=data.get("max_values"),
            disabled=data.get("disabled", False),
        )


class ButtonStyle(IntEnum):
    blue = 1
    gray = 2
    grey = 2
    green = 3
    red = 4
    URL = 5


class Button(Component):
    __slots__ = ("_style", "_label", "_id", "_url", "_disabled", "_emoji")

    def __init__(
        self,
        *,
        label: str = None,
        style: int = ButtonStyle.gray,
        id: str = None,
        custom_id: str = None,
        url: str = None,
        disabled: bool = False,
        emoji: Union[Emoji, PartialEmoji, str] = None,
    ):

        self._style = style
        self._label = label
        self._url = url
        self._disabled = disabled

        if emoji is not None:
            self._emoji = _get_partial_emoji(emoji)
        else:
            self._emoji = None

        if not self.style == ButtonStyle.URL:
            self._id = id or custom_id or str(uuid1())
        else:
            self._id = None

    def to_dict(self) -> dict:
        data = {
            "type": 2,
            "style": self.style,
            "label": self.label,
            "custom_id": self.id,
            "url": self.url if self.style == ButtonStyle.URL else None,
            "disabled": self.disabled,
        }
        if self.emoji:
            data["emoji"] = self.emoji.to_dict()
        return data

    @property
    def style(self) -> int:
        return self._style

    @property
    def label(self) -> str:
        return self._label

    @property
    def id(self) -> str:
        return self._id

    @property
    def custom_id(self) -> str:
        return self._id

    @property
    def url(self) -> Optional[str]:
        return self._url

    @property
    def disabled(self) -> bool:
        return self._disabled

    @property
    def emoji(self) -> PartialEmoji:
        return self._emoji

    @style.setter
    def style(self, value: int):
        if value == ButtonStyle.URL and self.id:
            raise InvalidArgument("Both ID and URL are set.")
        if not (1 <= value <= ButtonStyle.URL):
            raise InvalidArgument(f"Style must be between 1, {ButtonStyle.URL}.")

        self._style = value

    @label.setter
    def label(self, value: str):
        if not value and not self.emoji:
            raise InvalidArgument("Label should not be empty.")

        self._label = value

    @url.setter
    def url(self, value: str):
        if value and self.style != ButtonStyle.URL:
            raise InvalidArgument("Button style is not URL. You shouldn't provide URL.")

        self._url = value

    @id.setter
    def id(self, value: str):
        if self.style == ButtonStyle.URL:
            raise InvalidArgument(
                "Button style is set to URL. You shouldn't provide ID."
            )

        self._id = value

    @custom_id.setter
    def custom_id(self, value: str):
        if self.style == ButtonStyle.URL:
            raise InvalidArgument(
                "Button style is set to URL. You shouldn't provide ID."
            )

        self._id = value

    @disabled.setter
    def disabled(self, value: bool):
        self._disabled = value

    @emoji.setter
    def emoji(self, emoji: Union[Emoji, PartialEmoji, str]):
        self._emoji = _get_partial_emoji(emoji)

    def set_style(self, value: int):
        self.style = value

    def set_label(self, value: int):
        self.label = value

    def set_url(self, value: int):
        self.url = value

    def set_id(self, value: str):
        self.id = value

    def set_custom_id(self, value: str):
        self.custom_id = value

    def set_disabled(self, value: bool):
        self.disabled = value

    def set_emoji(self, emoji: Union[Emoji, PartialEmoji, str]):
        self.emoji = emoji

    @classmethod
    def from_json(cls, data: dict):
        emoji = data.get("emoji")
        return cls(
            style=data.get("style"),
            label=data.get("label"),
            id=data.get("custom_id"),
            url=data.get("url"),
            disabled=data.get("disabled", False),
            emoji=PartialEmoji(
                name=emoji["name"],
                animated=emoji.get("animated", False),
                id=emoji.get("id"),
            )
            if emoji
            else None,
        )


class ActionRow(Component):
    __slots__ = ("_components",)

    def __init__(self, *args: List[Component]):
        self._components = list(args) if args is not None else []

    def disable_components(self) -> List[Component]:
        def disable(component: Component):
            component.disabled = True
            return component

        self._components = list(map(disable, self._components))
        return self

    def __list__(self) -> List[Component]:
        return self.components

    def __len__(self) -> int:
        return len(self.components)

    def __iter__(self) -> Iterable[Component]:
        return iter(self.components)

    def __getitem__(self, index: int) -> Component:
        return self.components[index]

    def __setitem__(self, index: int, value: Component):
        self._components[index] = value

    def __delitem__(self, index: int):
        del self._components[index]

    def to_dict(self) -> dict:
        data = {
            "type": 1,
            "components": [component.to_dict() for component in self.components],
        }
        return data

    def append(self, component: Component):
        self.components.append(component)

    @property
    def components(self) -> List[Component]:
        return self._components

    @components.setter
    def components(self, value: List[Component]):
        self._components = value

    def set_components(self, value: List[Component]):
        self.components = value

    def add_component(self, value: Component):
        self.components.append(value)

    @classmethod
    def from_json(cls, data: dict):
        return cls(
            *[Button.from_json(component) for component in data.get("components")]
        )


def _get_component_type(type: int):
    return {1: ActionRow, 2: Button, 3: Select}[type]
